/*
**                     Copyright 1995-97 by KVASER AB
**         P.O Box 4076 S-51104 KINNAHULT, SWEDEN. Tel. +46 320 15287
**
** This software is furnished under a license and may be used and copied
** only in accordance with the terms of such license.
**
** Ver  Date    Sign    What
** 1.1  970323  DB      Spruced up, on popular demand.
** 1.2  970513  DB      Added the -p, -q, -i and -L flags.

*/
/*
** Test program for the PC-CAN card with 82527 controllers.
** Usage:
**   test527 [flags]
**
** The program will try to find the PC-CAN card.  If the card is found
** and the first Intel chip is a 82527, a message will be transmitted
** (at 500 kbit/s) from each one of the 14 possible transmit buffers. The
** message will be received by the second 82527.  The test is then performed
** again in the reverse order.
**
** This program was successfully compiled with Borland C++ 4.5 and run
** on a Pentium 66MHz with "enhanced ISA bus timing".
**
*/
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <string.h>
#include <bios.h>
#include <setjmp.h>

#include <i82527.h>
#include <pccan.h>

#define TRUE    1
#define FALSE   0

unsigned short Xilinx, Intel1, Intel2, Philips1, Philips2;
unsigned int Prescaler = 1;
int Interactive = 1;
int Verbose = 1;
int Loopmode = 0;
int IgnoreErrors = 0;
jmp_buf JmpBuf;

#define TEST_ID 530

// Currently, we don't use interrupts.
void interrupt (*oldIntHandler)();
int Irq = 5;

void interrupt CAN_IntHandler(void);

/*
** A printf with settable verbosity.
*/
int PRINTF(char *fmt, ...)
{
   va_list argptr;
   int cnt;

   if (!Verbose) return 0;
   va_start(argptr, fmt);
   cnt = vprintf(fmt, argptr);
   va_end(argptr);

   return(cnt);
}

// Delay for 100 ms.
void shortDelay( void)
{
    unsigned long t0, t1;

    t0 = biostime(0, 0L);
    do {
        t1 = biostime(0, 0L);
    } while ((t1-t0) < 2);
}

// Ctrl-Break handler.
int CtrlBreakHandler(void)
{
    longjmp(JmpBuf, 1);
    return 1;
}

/*
** Setup the PCCAN card.
*/
void setupCard(void)
{
    int i;
    // Select the on-board clock
    outportb(Xilinx + XREG_CLOCK, XREG_CLOCK_ONBOARD);

    // Reset the CAN controllers
    outportb(Xilinx + XREG_CTRL, 0);
    for (i = 0; i<1000; i++) {
        asm nop
    }
    outportb(Xilinx + XREG_CTRL, XREG_CTRL_NORESET);
}

void killCard(void)
{
    // Reset all controllers and leave it that way.
    outportb(Xilinx + XREG_CTRL, 0);

    //setvect(Irq, oldIntHandler);
}

/* =======================================================================
 * Initializes the 82527.
 */
void setup(int ctrl)
{
    int i;

    if (inportb(ctrl + CAN_CPUIR) & CPUIR_RSTST) {
        PRINTF( "Hardware reset is stuck in active state!\n");
        exit(1);
    }

    //Set DSC (makes the system clock SCLK = XTAL/2),
    //      DMC (makes the memory clock MCLK = SCLK/2)
    // This gives SCLK = 8MHz and MCLK = 4MHz

    outportb(ctrl + CAN_CPUIR, CPUIR_DSC|CPUIR_DMC);


    /* Initialize the 82527: */
    outportb(ctrl + CAN_CTRL, CTRL_INIT|CTRL_CCE);

    outportb(ctrl + CAN_STDGBLMASK0, 255);
    outportb(ctrl + CAN_STDGBLMASK1, 255);

    /* Bus configuration. */
    outportb(ctrl + CAN_BUSCON, BUSCON_COBY|BUSCON_DCT1|BUSCON_DCR1);

    /* The bit timing. SCLK = 8 MHz.
      The time quanta is set to 1/8 us  -  BRP=0.
      TSEG1=4
      TSEG2=1
    */
    outportb(ctrl + CAN_BTR0, 0x40|(Prescaler-1));  /* 01 xxxxxx */
    outportb(ctrl + CAN_BTR1, 0x41); /* 01 000001     Sample at position 5. */


    /* Set all message buffers to Invalid and Interrupts disable. */
    for (i = 1; i <= 15; i++) {
        int port = CANmsg(ctrl, i) + CANmsg_CTRL0;
        outportb(port, CLR(MsgVal) & CLR(TxIE) & CLR(RxIE));
    }

    outportb(ctrl + CAN_STAT, 0);

    /* Initialize interrupts. */
    /* And clear the init-bit and prevent changes of the configuration. */
    //oldIntHandler = getvect(Irq);
    //setvect(Irq, CAN_IntHandler);
    //outportb(ctrl + CAN_CTRL, CTRL_EIE | CTRL_IE);
    outportb(ctrl + CAN_CTRL, 0);
}


void interrupt CAN_IntHandler(void)
{
    //....
}


/* Among other things, disables interrupts.
 */
void shutdown(int ctrl)
{
    int i;

    outportb(ctrl + CAN_CTRL, 0);

    /* Set all message buffers to Invalid and Interrupts disable. */
    for (i = 1; i <= 15; i++) {
        outportb(CANmsg(ctrl, i) + CANmsg_CTRL0,
                    CLR(MsgVal) & CLR(TxIE) & CLR(RxIE));
    }

}

void testSend(int ctrl, int buf)
{
    int i;

    /* Send a message */
    /* Prevent the 82527 to transmit the message while we are updating it. */
    outportb(CANmsg(ctrl, buf) + CANmsg_CTRL0,
                CLR(TxIE) & CLR(RxIE) & CLR(IntPnd));
    outportb(CANmsg(ctrl, buf) + CANmsg_CTRL1,
                CLR(RmtPnd) & CLR(TxRqst) &  SET(CPUUpd));

    /* Set the message id to TEST_ID and up */
    outportb(CANmsg(ctrl, buf) + CANmsg_ARB,      (TEST_ID+buf)>>3);
    outportb(CANmsg(ctrl, buf) + CANmsg_ARB + 1, ((TEST_ID+buf)&7)<<5);

    outportb(CANmsg(ctrl, buf) + CANmsg_CTRL1, CLR(NewDat));


    /* The message should be 8 bytes, transmit, 11 bit id: */
    outportb(CANmsg(ctrl, buf) + CANmsg_CONF,
                CANmsg_DLC(8) | CANmsg_Standard | CANmsg_Transmit);

    outportb(CANmsg(ctrl, buf) + CANmsg_CTRL0, SET(MsgVal));
    outportb(CANmsg(ctrl, buf) + CANmsg_CTRL1, SET(NewDat) & SET(CPUUpd));

    /* Write the data bytes: */
    outportb(CANmsg(ctrl, buf) + CANmsg_DATA+0, buf);
    for (i = 1; i < 8; i++) {
        outportb(CANmsg(ctrl, buf) + CANmsg_DATA+i, i);
    }

    /* Transmit the message */
    outportb(CANmsg(ctrl, buf) + CANmsg_CTRL0, SET(MsgVal));
    outportb(CANmsg(ctrl, buf) + CANmsg_CTRL1, CLR(CPUUpd) & SET(TxRqst));
}

int checkTransmit(int ctrl, int buf)
{
    if ((inportb(CANmsg(ctrl, buf)+ CANmsg_CTRL1) & TEST(TxRqst)) == TEST(TxRqst)) {
        return FALSE;
    } else {
        return TRUE;
    }
}

void setupForReceive(int ctrl, int buf)
{
    outportb(CANmsg(ctrl, buf) + CANmsg_CTRL0,
                CLR(MsgVal) & CLR(TxIE) & CLR(RxIE) & CLR(IntPnd));

    outportb(CANmsg(ctrl, buf) + CANmsg_CTRL1,
                CLR(RmtPnd) & CLR(TxRqst) & CLR(NewDat));

    /* Set the message id to TEST_ID and up */
    outportb(CANmsg(ctrl, buf) + CANmsg_ARB,      (TEST_ID+buf)>>3);
    outportb(CANmsg(ctrl, buf) + CANmsg_ARB + 1, ((TEST_ID+buf)&7)<<5);

    /* The message should be 8 bytes, receive, 11 bit id: */
    outportb(CANmsg(ctrl, buf) + CANmsg_CONF,
                CANmsg_DLC(8) | CANmsg_Standard | CANmsg_Receive);

    outportb(CANmsg(ctrl, buf) + CANmsg_CTRL0, SET(MsgVal));
}


/* Check that a message was received in buffer buf*/
int testReceive(int ctrl, int buf)
{
    //return inportb(CANmsg(ctrl, buf) + CANmsg_DATA+0);
    if (inportb(CANmsg(ctrl, buf) + CANmsg_CTRL1) & TEST(NewDat)) {
        return TRUE;
    } else {
        return FALSE;
    }
}

// This routine is by no means water-proof, but it will delect a PC-CAN
// card if the circuits are newly resetted.
int CheckForCard(void)
{
    int addr = 0;
    static int addresses[] = { 0x300, 0x340, 0x200, 0x240, 0 };
    int i = 0;

    while (addresses[i] != 0) {
        if (inportb(addresses[i] + PHILIPS_OFFSET) == 0x21) {
            addr = addresses[i];
            Intel1 = INTEL_OFFSET + addr;
            Intel2 = INTEL_OFFSET + 0x1000 + addr;
            Philips1 = PHILIPS_OFFSET + addr;
            Philips2 = PHILIPS_OFFSET + 0x1000 + addr;
            Xilinx = XILINX_OFFSET + addr;
            break;
        }
        i++;
    }
    return addr;
}

void PrintCtrlType(int addr)
{
    int a = inportb(addr + 2) & 0x7F;
    int b = inportb(addr);
    if (a == 0x61) {
        PRINTF( "Intel 82527 at 0x%X.\n", addr);
    } else if (b == 0x21) {
        PRINTF( "Philips 82C200 at 0x%X.\n", addr);
    } else if (b != 0xff) {
        PRINTF( "Intel 82526 at 0x%X.\n", addr);
    } else {
        PRINTF( "Nothing at 0x%X.\n", addr);
    }
}


void PrintConfig(void)
{
    int addr = CheckForCard();

    if (!addr) {
        PRINTF( "No PC-CAN card was found!\n");
        exit(1);
    }
    PRINTF( "PC-CAN card detected at base address 0x%X.\n", addr);
    PrintCtrlType(INTEL_OFFSET + addr);
    PrintCtrlType(INTEL_OFFSET + 0x1000 + addr);
    PrintCtrlType(PHILIPS_OFFSET + addr);
    PrintCtrlType(PHILIPS_OFFSET + 0x1000 + addr);
}

void Usage(void)
{
    fprintf(stderr, "Usage: test527 [flags]\n");
    fprintf(stderr, "   -pXXX   Use XXX as prescaler value.\n");
    fprintf(stderr, "           1=1000 kbps, 2=500, 4=250 etc.\n");
    fprintf(stderr, "   -q      Quiet mode.\n");
    fprintf(stderr, "   -L      Loop mode.\n");
    fprintf(stderr, "   -i      Ignore errors.\n");
    exit(1);
}

int makeTest(void)
{
    int i;
    int stat = 0;
    
    setupCard();
    setup(Intel1);
    setup(Intel2);
    PRINTF("\n82527 #1: sending from buffer ");
    for (i=1; i<=14; i++) {
        PRINTF("%d..", i);
        setupForReceive(Intel2, i);
        testSend(Intel1, i);
        if (Interactive) (void)kbhit();
        shortDelay();
        if (!checkTransmit(Intel1, i)) {
            PRINTF("Tx failed ");
            if (!IgnoreErrors) {
                stat = 1;
                goto Goodbye;
            }
        }
        if (!testReceive(Intel2, i)) {
            PRINTF("Rx failed ");
            if (!IgnoreErrors) {
                stat = 1;
                goto Goodbye;
            }
        }
    }
    if (Verbose) putchar('\n');
    shutdown(Intel1);
    shutdown(Intel2);

    setup(Intel1);
    setup(Intel2);
    PRINTF("\82527 #2: sending from buffer ");
    for (i=1; i<=14; i++) {
        PRINTF("%d..", i);
        setupForReceive(Intel1, i);
        testSend(Intel2, i);
        if (Interactive) (void)kbhit();
        shortDelay();
        if (!checkTransmit(Intel2, i)) {
            PRINTF("Tx failed ");
            if (!IgnoreErrors) {
                stat = 1;
                goto Goodbye;
            }
        }
        if (!testReceive(Intel1, i)) {
            PRINTF("Rx failed ");
            if (!IgnoreErrors) {
                stat = 1;
                goto Goodbye;
            }
        }
    }
    if (Verbose) putchar('\n');

Goodbye:
    shutdown(Intel1);
    shutdown(Intel2);

    killCard();

    return stat;
}

void main(int argc, char *argv[])
{
    int i;
    int c;
    int stat;

    for (i=1; i<argc; i++) {
        char z;
        int tmp;
        if (strncmp(argv[i], "-p", 2) == 0) {
            if (sscanf(argv[i], "-p%d%c", &tmp, &z) == 1) {
                Prescaler = tmp & 63;
                if (Prescaler == 0) Prescaler = 1;
            } else {
                Usage();
            }
        } else if (strcmp(argv[i], "-q") == 0) {
            Interactive = 0;
            Verbose = 0;
        } else if (strcmp(argv[i], "-L") == 0) {
            Loopmode = 1;
        } else if (strcmp(argv[i], "-i") == 0) {
            IgnoreErrors = 1;
        } else {
            Usage();
        }
    }
    
    PRINTF("This program will send messages from buffers 1-14 with\n");
    PRINTF("identifiers %d+<buf no>. The bit rate will be %d kbit/s.\n\n",
              TEST_ID, 1000/Prescaler);
    PrintConfig();
    PRINTF("\nUsing what I assume are 82527's at addresses 0x%x and 0x%x.\n",
             Intel1, Intel2);
    if (Interactive) {
        printf("Press SPACE to start, or any other key to exit:");
        while (!kbhit()) /*Loop*/;
        c = getch();
        if (c != ' ') {
            putchar('\n');
            exit(0);
        }
    }

    ctrlbrk(CtrlBreakHandler);
    if (setjmp(JmpBuf) == 0) {
        do {
            stat = makeTest();
        } while (!stat && Loopmode);
    } else {
        stat = 1;
    }

    killCard();
    exit(stat);
}
